# frozen_string_literal: true

require 'builder'
require 'cucumber/formatter/backtrace_filter'
require 'cucumber/formatter/io'
require 'cucumber/formatter/interceptor'
require 'fileutils'
require 'json'

require "#{File.dirname(__FILE__)}/testwise_listener.rb"
require "#{File.dirname(__FILE__)}/testwise_runtime_support.rb"

class TestWiseCucumberFormatter
    include ::Cucumber::Formatter::Io
    
    include TestWiseRuntimeSupport
    
    class UnNamedFeatureError < StandardError
        def initialize(feature_file)
            super("The feature in '#{feature_file}' does not have a name. The JUnit XML format requires a name for the testsuite element.")
        end
    end
    
    def initialize(config)
        
        config.on_event :test_case_started, &method(:on_test_case_started)
        config.on_event :test_case_finished, &method(:on_test_case_finished)
        config.on_event :test_step_finished, &method(:on_test_step_finished)
        config.on_event :test_run_finished, &method(:on_test_run_finished)
        
        @config = config
        @features_data = Hash.new do |h, k|
            h[k] = {
                feature: nil,
                failures: 0,
                errors: 0,
                tests: 0,
                skipped: 0,
                time: 0,
                builder: Builder::XmlMarkup.new(:indent => 2)
            }
        end
        
        ::TestWise::Runtime.load_webdriver_support
        ::TestWise::Runtime.load_watir_support
        
    end
    
    def on_test_case_started(event)
        test_case = event.test_case
        test_case_name = NameBuilder.new(test_case)
        scenario = test_case_name.scenario_name

        
        hash = {}
        hash[:file_path] = test_case.location #  /Users/zhimin/work/projects/testwise-cpp/samples/demo-page-objects/features/01_login.feature:12
        hash[:test_case] = scenario
        json_data = hash.to_json
        
        connect_to_testwise("TESTSTART",  json_data + "\r\n")
        
        unless same_feature_as_previous_test_case?(test_case.feature)
            start_feature(test_case.feature)
        end
        @failing_step_source = nil
        
        # In order to fill out <system-err/> and <system-out/>, we need to
        # intercept the $stderr and $stdout
        @interceptedout = ::Cucumber::Formatter::Interceptor::Pipe.wrap(:stdout)
        @interceptederr = ::Cucumber::Formatter::Interceptor::Pipe.wrap(:stderr)
    end
    
    def on_test_step_finished(event)
        test_step, result = *event.attributes
        # puts("[DEBUG test step finished: " + test_step.inspect + " => " + result.inspect )
        return if @failing_step_source
        
        @failing_step_source = test_step.source.last unless result.ok?(@config.strict)
        
        ::Cucumber::Formatter::Interceptor::Pipe.unwrap! :stdout
        ::Cucumber::Formatter::Interceptor::Pipe.unwrap! :stderr
    end
    
    def on_test_case_finished(event)
        test_case, result = *event.attributes
        result = result.with_filtered_backtrace(Cucumber::Formatter::BacktraceFilter)
        test_case_name = NameBuilder.new(test_case)
        scenario = test_case_name.scenario_name
        scenario_designation = "#{scenario}#{test_case_name.name_suffix}"
        
        hash = {}
        hash[:file_path] = test_case.location #  /Users/zhimin/work/projects/testwise-cpp/samples/demo-page-objects/features/01_login.feature:12
        hash[:test_case] = scenario
        json_data = hash.to_json

        if result.class.name =~ /Passed/
            connect_to_testwise(" TESTPASS",  json_data + "\r\n")
        else
            # puts result.exception.inspect
            # puts result.exception.backtrace.inspect
            error_message = result.exception.message
            error_backtrace = result.exception.backtrace
            
            # prevent too big
            if error_message.size > 2000
                error_message = error_message[0..1744] + "..." + error_message[-256..-1]
            end
            hash[:error_message] = error_message
            
            if error_backtrace.size > 2000
                error_backtrace = error_backtrace[0..1000] + "..." + error_backtrace[-1000..-1]
            end
            hash[:error_backtrace] = error_backtrace
            
            json_data = hash.to_json
            connect_to_testwise(" TESTFAIL",  json_data + "\r\n")
        end

        
        output = create_output_string(test_case, scenario, result, test_case_name.row_name)
    end
    
    def on_test_run_finished(_event)
        @features_data.each { |file, data| end_feature(data) }
    end
    
    private
    
    def same_feature_as_previous_test_case?(feature)
        @current_feature_data && @current_feature_data[:feature].file == feature.file && @current_feature_data[:feature].location == feature.location
    end
    
    def start_feature(feature)
        raise UnNamedFeatureError.new(feature.file) if feature.name.empty?
        @current_feature_data = @features_data[feature.file]
        @current_feature_data[:feature] = feature unless @current_feature_data[:feature]
    end
    
    def end_feature(feature_data)
        @testsuite = Builder::XmlMarkup.new(:indent => 2)
        @testsuite.instruct!
        @testsuite.testsuite(
                             :failures => feature_data[:failures],
                             :errors => feature_data[:errors],
                             :skipped => feature_data[:skipped],
                             :tests => feature_data[:tests],
                             :time => format('%.6f', feature_data[:time]),
                             :name => feature_data[:feature].name
                             ) do
            @testsuite << feature_data[:builder].target!
        end
                             
    end
    
    def create_output_string(test_case, scenario, result, row_name)
        output = "#{test_case.keyword}: #{scenario}\n\n"
        return output if result.ok?(@config.strict)
        if test_case.keyword == 'Scenario'
            if @failing_step_source
                output += @failing_step_source.keyword.to_s unless hook?(@failing_step_source)
                output += "#{@failing_step_source}\n"
                else # An Around hook has failed
                output += "Around hook\n"
            end
            else
            output += "Example row: #{row_name}\n"
        end
        output + "\nMessage:\n"
    end
    
    def hook?(step)
        ['Before hook', 'After hook', 'AfterStep hook'].include? step.text
    end
    
    def build_testcase(result, scenario_designation, output)
        duration = ResultBuilder.new(result).test_case_duration
        @current_feature_data[:time] += duration
        classname = @current_feature_data[:feature].name
        name = scenario_designation
        
        @current_feature_data[:builder].testcase(:classname => classname, :name => name, :time => format('%.6f', duration)) do
            if !result.passed? && result.ok?(@config.strict)
                @current_feature_data[:builder].skipped
                @current_feature_data[:skipped] += 1
                elsif !result.passed?
                status = result.to_sym
                exception = get_backtrace_object(result)
                @current_feature_data[:builder].failure(:message => "#{status} #{name}", :type => status) do
                    @current_feature_data[:builder].cdata! output
                    @current_feature_data[:builder].cdata!(format_exception(exception)) if exception
                end
                @current_feature_data[:failures] += 1
            end
            @current_feature_data[:builder].tag!('system-out') do
                @current_feature_data[:builder].cdata! strip_control_chars(@interceptedout.buffer_string)
            end
            @current_feature_data[:builder].tag!('system-err') do
                @current_feature_data[:builder].cdata! strip_control_chars(@interceptederr.buffer_string)
            end
        end
        @current_feature_data[:tests] += 1
    end
    
    def get_backtrace_object(result)
        if result.failed?
            return result.exception
            elsif result.backtrace
            return result
            else
            return nil
        end
    end
    
    def format_exception(exception)
        (["#{exception.message} (#{exception.class})"] + exception.backtrace).join("\n")
    end
    
    def basename(feature_file)
        File.basename(feature_file.gsub(/[\\\/]/, '-'), '.feature')
    end
    
    def write_file(feature_filename, data)
        File.open(feature_filename, 'w') { |file| file.write(data) }
    end
    
    # strip control chars from cdata, to make it safe for external parsers
    def strip_control_chars(cdata)
        cdata.scan(/[[:print:]\t\n\r]/).join
    end
end

class NameBuilder
    attr_reader :scenario_name, :name_suffix, :row_name
    
    def initialize(test_case)
        @name_suffix = ''
        @row_name = ''
        test_case.describe_source_to self
    end
    
    def feature(*)
        self
    end
    
    def scenario(scenario)
        @scenario_name = (scenario.name.nil? || scenario.name == '') ? 'Unnamed scenario' : scenario.name
        self
    end
    
    def scenario_outline(outline)
        @scenario_name = (outline.name.nil? || outline.name == '') ? 'Unnamed scenario outline' : outline.name
        self
    end
    
    def examples_table(*)
        self
    end
    
    def examples_table_row(row)
        @row_name = '| ' + row.values.join(' | ') + ' |'
        @name_suffix = " (outline example : #{@row_name})"
        self
    end
end

class ResultBuilder
    attr_reader :test_case_duration
    def initialize(result)
        @test_case_duration = 0
        result.describe_to(self)
    end
    
    def passed(*) end
    
    def failed(*) end
    
    def undefined(*) end
    
    def skipped(*) end
    
    def pending(*) end
    
    def exception(*) end
    
    def duration(duration, *)
        duration.tap { |duration| @test_case_duration = duration.nanoseconds / 10**9.0 }
    end
end

=begin

require 'cucumber/formatter/console'
require 'cucumber/formatter/io'
require 'cucumber/formatter/duration_extractor'
require 'cucumber/formatter/hook_query_visitor'
require 'json'


# WARNING: The formatter TestWiseCucumberFormatter is using the deprecated formatter API which will be removed in v4.0 of Cucumber.

class TestWiseCucumberFormatter
  
  include ::Cucumber::Formatter::Console
  include ::Cucumber::Formatter::Io
  
  include TestWiseRuntimeSupport
  
  attr_reader :runtime

  def initialize(runtime, path_or_io, options)
    # NOTE: in new version of Cucumber:Formatter:IO the ensure_io takes only one argument    
    @runtime, @io, @options = runtime, ensure_io(path_or_io), options
    @previous_step_keyword = nil
    @snippets_input = []
    @total_duration = 0
    
    ::TestWise::Runtime.load_webdriver_support
    ::TestWise::Runtime.load_watir_support 
    
  end

  def before_feature(feature)
    hash = {}
    hash[:group_name] = feature.name
    if ENV["TESTWISE_RUNNING_AS"] == "test_case"
      # if run individual, the environable varible set
      hash[:test_case_count] = 1
    else
      hash[:test_case_count] =  feature.feature_elements.count
    end
    json_data = hash.to_json
    connect_to_testwise("    START",  json_data + "\r\n")    
  end
    
  def before_test_case(_test_case)
    unless @profile_information_printed
      print_profile_information
      @profile_information_printed = true
    end
    @previous_step_keyword = nil
    
    hash = {}
    hash[:file_path] = " "
    hash[:test_case] = _test_case.name
    json_data = hash.to_json
    connect_to_testwise("TESTSTART",  json_data + "\r\n")  	
  end

  def after_test_step(test_step, result)
    progress(result.to_sym) if !::Cucumber::Formatter::HookQueryVisitor.new(test_step).hook? || result.failed?
    collect_snippet_data(test_step, result) unless ::Cucumber::Formatter::HookQueryVisitor.new(test_step).hook?    
    # hash = {}
    # hash[:test_step] = test_step.name
    # json_data = hash.to_json

    test_step = test_step.inspect
    if test_step && test_step.strip =~ /#<Cucumber::Core::Test::Step:\s*(.*)>/ 
      test_step = $1
      if test_step.include?("/cucumber/filters")
        # ignore
        return
      end
      
      if test_step.include?(".feature")
        connect_to_testwise(" TRACE",  test_step)    
      end
    end  
    
    # new cucumber version with step prefix
  end

  def after_test_case(_test_case, result)
    @total_duration += ::Cucumber::Formatter::DurationExtractor.new(result).result_duration
    # puts "TestCase: #{_test_case.name}, #{_test_case.source.first}, result: #{result.class}"
    #test_case_file_path =  notification.example.metadata[:file_path]
    test_case_name = _test_case.name
    hash = {}
    hash[:file_path] = ""
    hash[:test_case] = test_case_name
    
    if result.class.name.include?("Passed")
      json_data = hash.to_json
      connect_to_testwise(" TESTPASS",  json_data + "\r\n")    
    else
      error_message = "   "
      error_backtrace = "   "
      
      # prevent too big
      if error_message.size > 2000
        error_message = error_message[0..1744] + "..." + error_message[-256..-1]
      end
      hash[:error_message] = error_message

      if error_backtrace.size > 2000
        error_backtrace = error_backtrace[0..1000] + "..." + error_backtrace[-1000..-1]
      end
      hash[:error_backtrace] = error_backtrace

      json_data = hash.to_json
      connect_to_testwise(" TESTFAIL",  json_data + "\r\n")
    end
    
  end

  def done
    @io.puts
    @io.puts
    print_summary
  end

  private

  def print_summary
    print_steps(:pending)
    print_steps(:failed)
    print_statistics(@total_duration, @options)
    print_snippets(@options)
    print_passing_wip(@options)
  end

  CHARS = {
    :passed    => '.',
    :failed    => 'F',
    :undefined => 'U',
    :pending   => 'P',
    :skipped   => '-'
  }

  def progress(status)
    char = CHARS[status]
    @io.print(format_string(char, status))
    @io.flush
  end

  def table_header_cell?(status)
    status == :skipped_param
  end
  
end    

=end
